# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Port
#

import pyomo.common.unittest as unittest
from io import StringIO

from pyomo.environ import (
    ConcreteModel,
    AbstractModel,
    Var,
    Set,
    NonNegativeReals,
    Binary,
    Reals,
    Integers,
    RangeSet,
)
from pyomo.network import Port, Arc


class TestPort(unittest.TestCase):
    def test_default_scalar_constructor(self):
        model = ConcreteModel()
        model.c = Port()
        self.assertEqual(len(model.c), 1)
        self.assertEqual(len(model.c.vars), 0)

        model = AbstractModel()
        model.c = Port()
        self.assertEqual(len(model.c), 0)
        # FIXME: Not sure I like this behavior: but since this is
        # (currently) an attribute, there is no way to check for
        # construction without converting it to a property.
        #
        # TODO: if we move away from multiple inheritance for
        # simplevars, then this can trigger an exception (cleanly)
        self.assertEqual(len(model.c.vars), 0)

        inst = model.create_instance()
        self.assertEqual(len(inst.c), 1)
        self.assertEqual(len(inst.c.vars), 0)

    def test_default_indexed_constructor(self):
        model = ConcreteModel()
        model.c = Port([1, 2, 3])
        self.assertEqual(len(model.c), 3)
        self.assertEqual(len(model.c[1].vars), 0)

        model = AbstractModel()
        model.c = Port([1, 2, 3])
        self.assertEqual(len(model.c), 0)
        self.assertRaises(ValueError, model.c.__getitem__, 1)

        inst = model.create_instance()
        self.assertEqual(len(inst.c), 3)
        self.assertEqual(len(inst.c[1].vars), 0)

    def test_add_scalar_vars(self):
        pipe = ConcreteModel()
        pipe.flow = Var()
        pipe.pIn = Var(within=NonNegativeReals)
        pipe.pOut = Var(within=NonNegativeReals)

        pipe.OUT = Port()
        pipe.OUT.add(pipe.flow, "flow")
        pipe.OUT.add(pipe.pOut, "pressure")
        self.assertEqual(len(pipe.OUT), 1)
        self.assertEqual(len(pipe.OUT.vars), 2)
        self.assertFalse(pipe.OUT.vars['flow'].is_expression_type())

        pipe.IN = Port()
        pipe.IN.add(-pipe.flow, "flow")
        pipe.IN.add(pipe.pIn, "pressure")
        self.assertEqual(len(pipe.IN), 1)
        self.assertEqual(len(pipe.IN.vars), 2)
        self.assertTrue(pipe.IN.vars['flow'].is_expression_type())

    def test_add_indexed_vars(self):
        pipe = ConcreteModel()
        pipe.SPECIES = Set(initialize=['a', 'b', 'c'])
        pipe.flow = Var()
        pipe.composition = Var(pipe.SPECIES)
        pipe.pIn = Var(within=NonNegativeReals)

        pipe.OUT = Port()
        pipe.OUT.add(pipe.flow, "flow")
        pipe.OUT.add(pipe.composition, "composition")
        pipe.OUT.add(pipe.pIn, "pressure")

        self.assertEqual(len(pipe.OUT), 1)
        self.assertEqual(len(pipe.OUT.vars), 3)

    def test_fixed(self):
        pipe = ConcreteModel()
        pipe.SPECIES = Set(initialize=['a', 'b', 'c'])
        pipe.flow = Var()
        pipe.composition = Var(pipe.SPECIES)
        pipe.pIn = Var(within=NonNegativeReals)

        pipe.OUT = Port()
        self.assertTrue(pipe.OUT.is_fixed())

        pipe.OUT.add(pipe.flow, "flow")
        self.assertFalse(pipe.OUT.is_fixed())

        pipe.flow.fix(0)
        self.assertTrue(pipe.OUT.is_fixed())

        pipe.OUT.add(-pipe.pIn, "pressure")
        self.assertFalse(pipe.OUT.is_fixed())

        pipe.pIn.fix(1)
        self.assertTrue(pipe.OUT.is_fixed())

        pipe.OUT.add(pipe.composition, "composition")
        self.assertFalse(pipe.OUT.is_fixed())

        pipe.composition['a'].fix(1)
        self.assertFalse(pipe.OUT.is_fixed())

        pipe.composition['b'].fix(1)
        pipe.composition['c'].fix(1)
        self.assertTrue(pipe.OUT.is_fixed())

        m = ConcreteModel()
        m.SPECIES = Set(initialize=['a', 'b', 'c'])
        m.flow = Var()
        m.composition = Var(m.SPECIES)
        m.pIn = Var(within=NonNegativeReals)

        m.port = Port()
        m.port.add(m.flow, "flow")
        m.port.add(-m.pIn, "pressure")
        m.port.add(m.composition, "composition")
        m.port.fix()
        self.assertTrue(m.port.is_fixed())

    def test_polynomial_degree(self):
        pipe = ConcreteModel()
        pipe.SPECIES = Set(initialize=['a', 'b', 'c'])
        pipe.flow = Var()
        pipe.composition = Var(pipe.SPECIES)
        pipe.pIn = Var(within=NonNegativeReals)

        pipe.OUT = Port()
        self.assertEqual(pipe.OUT.polynomial_degree(), 0)

        pipe.OUT.add(pipe.flow, "flow")
        self.assertEqual(pipe.OUT.polynomial_degree(), 1)

        pipe.flow.fix(0)
        self.assertEqual(pipe.OUT.polynomial_degree(), 0)

        pipe.OUT.add(-pipe.pIn, "pressure")
        self.assertEqual(pipe.OUT.polynomial_degree(), 1)

        pipe.pIn.fix(1)
        self.assertEqual(pipe.OUT.polynomial_degree(), 0)

        pipe.OUT.add(pipe.composition, "composition")
        self.assertEqual(pipe.OUT.polynomial_degree(), 1)

        pipe.composition['a'].fix(1)
        self.assertEqual(pipe.OUT.polynomial_degree(), 1)

        pipe.composition['b'].fix(1)
        pipe.composition['c'].fix(1)
        self.assertEqual(pipe.OUT.polynomial_degree(), 0)

        pipe.OUT.add(pipe.flow * pipe.pIn, "quadratic")
        self.assertEqual(pipe.OUT.polynomial_degree(), 0)

        pipe.flow.unfix()
        self.assertEqual(pipe.OUT.polynomial_degree(), 1)

        pipe.pIn.unfix()
        self.assertEqual(pipe.OUT.polynomial_degree(), 2)

        pipe.OUT.add(pipe.flow / pipe.pIn, "nonLin")
        self.assertEqual(pipe.OUT.polynomial_degree(), None)

    def test_potentially_variable(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Port()
        self.assertTrue(m.p.is_potentially_variable())
        m.p.add(-m.x)
        self.assertTrue(m.p.is_potentially_variable())

    def test_binary(self):
        m = ConcreteModel()
        m.x = Var(domain=Binary)
        m.y = Var(domain=Reals)
        m.p = Port()

        self.assertTrue(m.p.is_binary())

        m.p.add(m.x)
        self.assertTrue(m.p.is_binary())

        m.p.add(-m.x, "foo")
        self.assertTrue(m.p.is_binary())

        m.p.add(m.y)
        self.assertFalse(m.p.is_binary())

        m.p.remove('y')
        self.assertTrue(m.p.is_binary())

        m.p.add(-m.y, "bar")
        self.assertFalse(m.p.is_binary())

    def test_integer(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers)
        m.y = Var(domain=Reals)
        m.p = Port()

        self.assertTrue(m.p.is_integer())

        m.p.add(m.x)
        self.assertTrue(m.p.is_integer())

        m.p.add(-m.x, "foo")
        self.assertTrue(m.p.is_integer())

        m.p.add(m.y)
        self.assertFalse(m.p.is_integer())

        m.p.remove('y')
        self.assertTrue(m.p.is_integer())

        m.p.add(-m.y, "bar")
        self.assertFalse(m.p.is_integer())

    def test_continuous(self):
        m = ConcreteModel()
        m.x = Var(domain=Reals)
        m.y = Var(domain=Integers)
        m.p = Port()

        self.assertTrue(m.p.is_continuous())

        m.p.add(m.x)
        self.assertTrue(m.p.is_continuous())

        m.p.add(-m.x, "foo")
        self.assertTrue(m.p.is_continuous())

        m.p.add(m.y)
        self.assertFalse(m.p.is_continuous())

        m.p.remove('y')
        self.assertTrue(m.p.is_continuous())

        m.p.add(-m.y, "bar")
        self.assertFalse(m.p.is_continuous())

    def test_getattr(self):
        m = ConcreteModel()
        m.x = Var()
        m.port = Port()
        m.port.add(m.x)
        self.assertIs(m.port.x, m.x)

    def test_arc_lists(self):
        m = ConcreteModel()
        m.x = Var()
        m.p1 = Port()
        m.p2 = Port()
        m.p3 = Port()
        m.p4 = Port()
        m.p5 = Port()
        m.p1.add(m.x)
        m.p2.add(m.x)
        m.p3.add(m.x)
        m.p4.add(m.x)
        m.p5.add(m.x)
        m.a1 = Arc(source=m.p1, destination=m.p2)
        m.a2 = Arc(source=m.p1, destination=m.p3)
        m.a3 = Arc(source=m.p4, destination=m.p1)
        m.a4 = Arc(source=m.p5, destination=m.p1)

        self.assertEqual(len(m.p1.dests()), 2)
        self.assertEqual(len(m.p1.sources()), 2)
        self.assertEqual(len(m.p1.arcs()), 4)
        self.assertEqual(len(m.p2.dests()), 0)
        self.assertEqual(len(m.p2.sources()), 1)
        self.assertEqual(len(m.p2.arcs()), 1)
        self.assertIn(m.a1, m.p1.dests())
        self.assertIn(m.a1, m.p2.sources())
        self.assertNotIn(m.a1, m.p1.sources())
        self.assertNotIn(m.a1, m.p2.dests())

        self.assertEqual(len(m.p1.dests(active=True)), 2)
        self.assertEqual(len(m.p1.sources(active=True)), 2)
        self.assertEqual(len(m.p1.arcs(active=True)), 4)
        self.assertEqual(len(m.p2.dests(active=True)), 0)
        self.assertEqual(len(m.p2.sources(active=True)), 1)
        self.assertEqual(len(m.p2.arcs(active=True)), 1)
        self.assertIn(m.a1, m.p1.dests(active=True))
        self.assertIn(m.a1, m.p2.sources(active=True))
        self.assertNotIn(m.a1, m.p1.sources(active=True))
        self.assertNotIn(m.a1, m.p2.dests(active=True))

        m.a2.deactivate()

        self.assertNotIn(m.a2, m.p1.dests(active=True))
        self.assertNotIn(m.a2, m.p3.sources(active=True))
        self.assertIn(m.a2, m.p1.dests(active=False))
        self.assertIn(m.a2, m.p3.sources(active=False))
        self.assertIn(m.a2, m.p1.arcs(active=False))
        self.assertIn(m.a2, m.p3.arcs(active=False))
        self.assertIn(m.a2, m.p1.dests())
        self.assertIn(m.a2, m.p3.sources())
        self.assertIn(m.a2, m.p1.arcs())
        self.assertIn(m.a2, m.p3.arcs())

    def test_remove(self):
        m = ConcreteModel()
        m.x = Var()
        m.port = Port()
        m.port.add(m.x)
        self.assertIn('x', m.port.vars)
        self.assertIn('x', m.port._rules)
        m.port.remove('x')
        self.assertNotIn('x', m.port.vars)
        self.assertNotIn('x', m.port._rules)

    def test_extends(self):
        m = ConcreteModel()
        m.x = Var()
        m.p1 = Port()
        m.p1.add(m.x, rule=Port.Extensive)
        m.p2 = Port(extends=m.p1)
        self.assertIs(m.p2.x, m.x)
        self.assertIs(m.p2.rule_for('x'), Port.Extensive)
        self.assertTrue(m.p2.is_extensive('x'))
        self.assertFalse(m.p2.is_equality('x'))

    def test_add_from_containers(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.p1 = Port(initialize=[m.x, m.y])
        m.p2 = Port(initialize=[(m.x, Port.Equality), (m.y, Port.Extensive)])
        m.p3 = Port(initialize=dict(this=m.x, that=m.y))
        m.p4 = Port(
            initialize=dict(this=(m.x, Port.Equality), that=(m.y, Port.Extensive))
        )

        self.assertIs(m.p1.x, m.x)
        self.assertIs(m.p1.y, m.y)
        self.assertIs(m.p2.x, m.x)
        self.assertTrue(m.p2.is_equality('x'))
        self.assertIs(m.p2.y, m.y)
        self.assertTrue(m.p2.is_extensive('y'))
        self.assertIs(m.p3.this, m.x)
        self.assertIs(m.p3.that, m.y)
        self.assertIs(m.p4.this, m.x)
        self.assertTrue(m.p4.is_equality('this'))
        self.assertIs(m.p4.that, m.y)
        self.assertTrue(m.p4.is_extensive('that'))

    def test_fix_unfix(self):
        m = ConcreteModel()
        m.x = Var()
        m.port = Port()
        m.port.add(m.x)
        m.x.value = 10
        m.port.fix()
        self.assertTrue(m.x.is_fixed())
        m.port.unfix()
        self.assertFalse(m.x.is_fixed())

    def test_iter_vars(self):
        def contains(item, container):
            # use this instead of "in" to avoid "==" operation
            return any(item is mem for mem in container)

        m = ConcreteModel()
        m.s = RangeSet(5)
        m.x = Var()
        m.y = Var()
        m.z = Var(m.s)
        m.a = Var()
        m.b = Var()
        m.c = Var()
        expr = m.a + m.b * m.c
        p = m.p = Port()

        v = list(p.iter_vars())
        self.assertEqual(len(v), 0)

        p.add(m.x)
        v = list(p.iter_vars())
        self.assertEqual(len(v), 1)

        p.add(m.y)
        v = list(p.iter_vars())
        self.assertEqual(len(v), 2)

        p.add(m.z)
        v = list(p.iter_vars())
        self.assertEqual(len(v), 7)
        self.assertTrue(contains(m.x, v))
        self.assertTrue(contains(m.z[3], v))

        p.add(expr, "expr")
        v = list(p.iter_vars())
        self.assertEqual(len(v), 8)
        self.assertTrue(contains(m.x, v))
        self.assertTrue(contains(m.z[3], v))
        self.assertTrue(contains(expr, v))

        m.x.fix(0)
        v = list(p.iter_vars())
        self.assertEqual(len(v), 8)

        v = list(p.iter_vars(fixed=False))
        self.assertEqual(len(v), 7)
        self.assertTrue(contains(m.z[3], v))
        self.assertTrue(contains(expr, v))

        v = list(p.iter_vars(fixed=True))
        self.assertEqual(len(v), 1)
        self.assertIn(m.x, v)

        v = list(p.iter_vars(expr_vars=True))
        self.assertEqual(len(v), 10)
        self.assertFalse(contains(expr, v))
        self.assertTrue(contains(m.a, v))
        self.assertTrue(contains(m.b, v))

        m.a.fix(0)
        v = list(p.iter_vars(expr_vars=True, fixed=False))
        self.assertEqual(len(v), 8)

        m.b.fix(0)
        m.c.fix(0)
        v = list(p.iter_vars(expr_vars=True, fixed=False))
        self.assertEqual(len(v), 6)

        v = list(p.iter_vars(fixed=False))
        self.assertEqual(len(v), 6)
        self.assertFalse(contains(expr, v))

        v = list(p.iter_vars(expr_vars=True, names=True))
        self.assertEqual(len(v), 10)
        self.assertEqual(len(v[0]), 3)
        for t in v:
            if t[0] == 'x':
                self.assertIs(t[2], m.x)
                break

    def test_pprint(self):
        pipe = ConcreteModel()
        pipe.SPECIES = Set(initialize=['a', 'b', 'c'])
        pipe.flow = Var()
        pipe.composition = Var(pipe.SPECIES)
        pipe.pIn = Var(within=NonNegativeReals)

        pipe.OUT = Port(implicit=['imp'])
        pipe.OUT.add(-pipe.flow, "flow")
        pipe.OUT.add(pipe.composition, "composition")
        pipe.OUT.add(pipe.composition['a'], "comp_a")
        pipe.OUT.add(pipe.pIn, "pressure")

        os = StringIO()
        pipe.OUT.pprint(ostream=os)
        self.assertEqual(
            os.getvalue(),
            """OUT : Size=1, Index=None
    Key  : Name        : Size : Variable
    None :      comp_a :    1 : composition[a]
         : composition :    3 : composition
         :        flow :    1 : - flow
         :         imp :    - : None
         :    pressure :    1 : pIn
""",
        )

        def _IN(m, i):
            return {'pressure': pipe.pIn, 'flow': pipe.composition[i] * pipe.flow}

        pipe.IN = Port(pipe.SPECIES, rule=_IN)
        os = StringIO()
        pipe.IN.pprint(ostream=os)
        self.assertEqual(
            os.getvalue(),
            """IN : Size=3, Index=SPECIES
    Key : Name     : Size : Variable
      a :     flow :    1 : composition[a]*flow
        : pressure :    1 :                 pIn
      b :     flow :    1 : composition[b]*flow
        : pressure :    1 :                 pIn
      c :     flow :    1 : composition[c]*flow
        : pressure :    1 :                 pIn
""",
        )

    def test_display(self):
        pipe = ConcreteModel()
        pipe.SPECIES = Set(initialize=['a', 'b', 'c'])
        pipe.flow = Var(initialize=10)
        pipe.composition = Var(pipe.SPECIES, initialize=lambda m, i: ord(i) - ord('a'))
        pipe.pIn = Var(within=NonNegativeReals, initialize=3.14)

        pipe.OUT = Port(implicit=['imp'])
        pipe.OUT.add(-pipe.flow, "flow")
        pipe.OUT.add(pipe.composition, "composition")
        pipe.OUT.add(pipe.pIn, "pressure")

        os = StringIO()
        pipe.OUT.display(ostream=os)
        self.assertEqual(
            os.getvalue(),
            """OUT : Size=1
    Key  : Name        : Value
    None : composition : {'a': 0, 'b': 1, 'c': 2}
         :        flow : -10
         :         imp : -
         :    pressure : 3.14
""",
        )

        def _IN(m, i):
            return {'pressure': pipe.pIn, 'flow': pipe.composition[i] * pipe.flow}

        pipe.IN = Port(pipe.SPECIES, rule=_IN)
        os = StringIO()
        pipe.IN.display(ostream=os)
        self.assertEqual(
            os.getvalue(),
            """IN : Size=3
    Key : Name     : Value
      a :     flow :     0
        : pressure :  3.14
      b :     flow :    10
        : pressure :  3.14
      c :     flow :    20
        : pressure :  3.14
""",
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from abc import ABCMeta, abstractmethod
import enum
from typing import Optional, Union, Tuple
from scipy.sparse import spmatrix
import numpy as np
from pyomo.contrib.pynumero.sparse.base_block import BaseBlockVector, BaseBlockMatrix
from pyomo.contrib.pynumero.sparse import BlockVector, BlockMatrix


class LinearSolverStatus(enum.Enum):
    successful = 0
    not_enough_memory = 1
    singular = 2
    error = 3
    warning = 4
    max_iter = 5


class LinearSolverResults:
    def __init__(self, status: Optional[LinearSolverStatus] = None):
        self.status = status


class LinearSolverInterface(object, metaclass=ABCMeta):
    @abstractmethod
    def solve(
        self,
        matrix: Union[spmatrix, BlockMatrix],
        rhs: Union[np.ndarray, BlockVector],
        raise_on_error: bool = True,
    ) -> Tuple[Optional[Union[np.ndarray, BlockVector]], LinearSolverResults]:
        pass


class DirectLinearSolverInterface(LinearSolverInterface, metaclass=ABCMeta):
    @abstractmethod
    def do_symbolic_factorization(
        self, matrix: Union[spmatrix, BlockMatrix], raise_on_error: bool = True
    ) -> LinearSolverResults:
        pass

    @abstractmethod
    def do_numeric_factorization(
        self, matrix: Union[spmatrix, BlockMatrix], raise_on_error: bool = True
    ) -> LinearSolverResults:
        pass

    @abstractmethod
    def do_back_solve(
        self, rhs: Union[np.ndarray, BlockVector], raise_on_error: bool = True
    ) -> Tuple[Optional[Union[np.ndarray, BlockVector]], LinearSolverResults]:
        pass

    def solve(
        self,
        matrix: Union[spmatrix, BlockMatrix],
        rhs: Union[np.ndarray, BlockVector],
        raise_on_error: bool = True,
    ) -> Tuple[Optional[Union[np.ndarray, BlockVector]], LinearSolverResults]:
        symbolic_res = self.do_symbolic_factorization(
            matrix, raise_on_error=raise_on_error
        )
        if symbolic_res.status != LinearSolverStatus.successful:
            return None, symbolic_res
        numeric_res = self.do_numeric_factorization(
            matrix, raise_on_error=raise_on_error
        )
        if numeric_res.status != LinearSolverStatus.successful:
            return None, numeric_res
        return self.do_back_solve(rhs, raise_on_error=raise_on_error)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import time
import pyomo.opt
from pyomo.opt import SolverFactory
from pyomo.core import TransformationFactory
from pyomo.common.collections import Bunch


@SolverFactory.register(
    'mpec_nlp', doc='MPEC solver that optimizes a nonlinear transformation'
)
class MPEC_Solver1(pyomo.opt.OptSolver):
    def __init__(self, **kwds):
        kwds['type'] = 'mpec_nlp'
        pyomo.opt.OptSolver.__init__(self, **kwds)
        self._metasolver = True

    def _presolve(self, *args, **kwds):
        #
        # Cache the instance
        #
        self._instance = args[0]
        pyomo.opt.OptSolver._presolve(self, *args, **kwds)

    def _apply_solver(self):
        start_time = time.time()
        #
        # Transform instance
        #
        xfrm = TransformationFactory('mpec.simple_nonlinear')
        xfrm.apply_to(self._instance)
        #
        # Solve with a specified solver
        #
        solver = self.options.solver
        if not self.options.solver:  # pragma:nocover
            self.options.solver = solver = 'ipopt'

        # use the with block here so that deactivation of the
        # solver plugin always occurs thereby avoiding memory
        # leaks caused by plugins!
        with pyomo.opt.SolverFactory(solver) as opt:
            self.results = []
            epsilon_final = self.options.get('epsilon_final', 1e-7)
            epsilon = self.options.get('epsilon_initial', epsilon_final)
            while True:
                self._instance.mpec_bound.set_value(epsilon)
                #
                # **NOTE: It would be better to override _presolve on the
                #         base class of this solver as you might be
                #         missing a number of keywords that were passed
                #         into the solve method (e.g., none of the
                #         io_options are getting relayed to the subsolver
                #         here).
                #
                res = opt.solve(
                    self._instance, tee=self._tee, timelimit=self._timelimit
                )
                self.results.append(res)
                epsilon /= 10.0
                if epsilon < epsilon_final:
                    break
            #
            # Reclassify the Complementarity components
            #
            from pyomo.mpec import Complementarity

            for cuid in self._instance._transformation_data[
                'mpec.simple_nonlinear'
            ].compl_cuids:
                cobj = cuid.find_component_on(self._instance)
                cobj.parent_block().reclassify_component_type(cobj, Complementarity)
            #
            # Update timing
            #
            stop_time = time.time()
            self.wall_time = stop_time - start_time
            #
            # Return the sub-solver return condition value and log
            #
            return Bunch(rc=getattr(opt, '_rc', None), log=getattr(opt, '_log', None))

    def _postsolve(self):
        #
        # Create a results object
        #
        results = pyomo.opt.SolverResults()
        #
        # SOLVER
        #
        solv = results.solver
        solv.name = self.options.subsolver
        solv.wallclock_time = self.wall_time
        cpu_ = []
        for res in self.results:  # pragma:nocover
            if not getattr(res.solver, 'cpu_time', None) is None:
                cpu_.append(res.solver.cpu_time)
        if len(cpu_) > 0:  # pragma:nocover
            solv.cpu_time = sum(cpu_)
        # solv.termination_condition = pyomo.opt.TerminationCondition.maxIterations
        #
        # PROBLEM
        #
        self._instance.compute_statistics()
        prob = results.problem
        prob.name = self._instance.name
        prob.number_of_constraints = self._instance.statistics.number_of_constraints
        prob.number_of_variables = self._instance.statistics.number_of_variables
        prob.number_of_binary_variables = (
            self._instance.statistics.number_of_binary_variables
        )
        prob.number_of_integer_variables = (
            self._instance.statistics.number_of_integer_variables
        )
        prob.number_of_continuous_variables = (
            self._instance.statistics.number_of_continuous_variables
        )
        prob.number_of_objectives = self._instance.statistics.number_of_objectives
        #
        # SOLUTION(S)
        #
        self._instance.solutions.store_to(results)
        #
        # Uncache the instance and return the results
        #
        self._instance = None
        return results



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import os

from pyomo.common.tempfiles import TempfileManager
from pyomo.opt.base import ProblemFormat
from pyomo.opt.base.convert import ProblemConverterFactory
from pyomo.solvers.plugins.converter.pico import PicoMIPConverter
from pyomo.core.kernel.block import IBlock


@ProblemConverterFactory.register('pyomo')
class PyomoMIPConverter:
    pico_converter = PicoMIPConverter()

    def can_convert(self, from_type, to_type):
        """Returns true if this object supports the specified conversion"""
        if from_type != ProblemFormat.pyomo:
            return False
        #
        # Return True for specific from/to pairs
        #
        if to_type in (
            ProblemFormat.nl,
            ProblemFormat.cpxlp,
            ProblemFormat.osil,
            ProblemFormat.bar,
            ProblemFormat.mps,
        ):
            return True

        return False

    def apply(self, *args, **kwds):
        """
        Generate a NL or LP file from Pyomo, and then do subsequent
        conversions.
        """

        import pyomo.scripting.convert

        capabilities = kwds.pop("capabilities", None)

        # all non-consumed keywords are assumed to be options
        # that should be passed to the writer.
        io_options = {}
        for kwd, value in kwds.items():
            io_options[kwd] = value
        kwds.clear()

        if isinstance(args[2], str):
            instance = None
        else:
            instance = args[2]

        if args[1] == ProblemFormat.cpxlp:
            problem_filename = TempfileManager.create_tempfile(suffix='.pyomo.lp')
            if instance is not None:
                if isinstance(instance, IBlock):
                    symbol_map_id = instance.write(
                        problem_filename,
                        format=ProblemFormat.cpxlp,
                        _solver_capability=capabilities,
                        _called_by_solver=True,
                        **io_options,
                    )
                else:
                    problem_filename, symbol_map_id = instance.write(
                        filename=problem_filename,
                        format=ProblemFormat.cpxlp,
                        solver_capability=capabilities,
                        io_options=io_options,
                    )
                return (problem_filename,), symbol_map_id
            else:
                #
                # I'm simply exposing a fatal issue with
                # this code path. How would we convert the
                # collected keywords into command-line
                # arguments that can be sent to the writer?
                #
                if len(io_options):
                    raise ValueError(
                        "The following io_options will be ignored "
                        "(please create a bug report):\n\t"
                        + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
                    )

                ans = pyomo.scripting.convert.pyomo2lp(
                    ['--output', problem_filename, args[2]]
                )
                if ans.errorcode:
                    raise RuntimeError(
                        "pyomo2lp conversion "
                        "returned nonzero error code "
                        "(%s)" % ans.errorcode
                    )

                model = ans.retval
                problem_filename = model.filename
                symbol_map = model.symbol_map
                return (problem_filename,), symbol_map

        elif args[1] == ProblemFormat.bar:
            problem_filename = TempfileManager.create_tempfile(suffix='.pyomo.bar')
            if instance is not None:
                if isinstance(instance, IBlock):
                    symbol_map_id = instance.write(
                        problem_filename,
                        format=ProblemFormat.bar,
                        _solver_capability=capabilities,
                        _called_by_solver=True,
                        **io_options,
                    )
                else:
                    problem_filename, symbol_map_id = instance.write(
                        filename=problem_filename,
                        format=ProblemFormat.bar,
                        solver_capability=capabilities,
                        io_options=io_options,
                    )
                return (problem_filename,), symbol_map_id
            else:
                #
                # I'm simply exposing a fatal issue with
                # this code path. How would we convert the
                # collected keywords into command-line
                # arguments that can be sent to the writer?
                #
                if len(io_options):
                    raise ValueError(
                        "The following io_options will be ignored "
                        "(please create a bug report):\n\t"
                        + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
                    )

                ans = pyomo.scripting.convert.pyomo2bar(
                    ['--output', problem_filename, args[2]]
                )
                if ans.errorcode:
                    raise RuntimeError(
                        "pyomo2bar conversion "
                        "returned nonzero error code "
                        "(%s)" % ans.errorcode
                    )
                model = ans.retval
                problem_filename = model.filename
                symbol_map = model.symbol_map
                return (problem_filename,), symbol_map

        elif args[1] in [ProblemFormat.mps, ProblemFormat.nl]:
            if args[1] == ProblemFormat.nl:
                problem_filename = TempfileManager.create_tempfile(suffix='.pyomo.nl')
                if io_options.get("symbolic_solver_labels", False):
                    TempfileManager.add_tempfile(
                        problem_filename[:-3] + ".row", exists=False
                    )
                    TempfileManager.add_tempfile(
                        problem_filename[:-3] + ".col", exists=False
                    )
            else:
                assert args[1] == ProblemFormat.mps
                problem_filename = TempfileManager.create_tempfile(suffix='.pyomo.mps')
            if instance is not None:
                if isinstance(instance, IBlock):
                    symbol_map_id = instance.write(
                        problem_filename,
                        format=args[1],
                        _solver_capability=capabilities,
                        _called_by_solver=True,
                        **io_options,
                    )
                else:
                    problem_filename, symbol_map_id = instance.write(
                        filename=problem_filename,
                        format=args[1],
                        solver_capability=capabilities,
                        io_options=io_options,
                    )
                return (problem_filename,), symbol_map_id
            else:
                #
                # I'm simply exposing a fatal issue with
                # this code path. How would we convert the
                # collected keywords into command-line
                # arguments that can be sent to the writer?
                #
                if len(io_options):
                    raise ValueError(
                        "The following io_options will be ignored "
                        "(please create a bug report):\n\t"
                        + "\n\t".join("%s = %s" % (k, v) for k, v in io_options.items())
                    )

                ans = pyomo.scripting.convert.pyomo2nl(
                    ['--output', problem_filename, args[2]]
                )
                if ans.errorcode:
                    raise RuntimeError(
                        "pyomo2nl conversion "
                        "returned nonzero error "
                        "code (%s)" % ans.errorcode
                    )
                model = ans.retval
                problem_filename = model.filename
                symbol_map = model.symbol_map

                if args[1] == ProblemFormat.nl:
                    return (problem_filename,), symbol_map
                #
                # Convert from NL to MPS
                #
                # TBD: We don't support a variable map file when going
                #      from NL to MPS within the PICO converter.
                # NOTE: this is a problem with the MPS writer that is
                #       provided by COIN-OR
                # NOTE: we should generalize this so it doesn't strictly
                #       depend on the PICO converter utility.
                #
                ans = self.pico_converter.apply(
                    ProblemFormat.nl, ProblemFormat.mps, problem_filename
                )
                os.remove(problem_filename)
                return ans

        elif args[1] == ProblemFormat.osil:
            if False:
                problem_filename = TempfileManager.create_tempfile(suffix='pyomo.osil')
                if instance:
                    if isinstance(instance, IBlock):
                        symbol_map_id = instance.write(
                            problem_filename,
                            format=ProblemFormat.osil,
                            _solver_capability=capabilities,
                            _called_by_solver=True,
                            **io_options,
                        )
                    else:
                        problem_filename, symbol_map_id = instance.write(
                            filename=problem_filename,
                            format=ProblemFormat.osil,
                            solver_capability=capabilities,
                            io_options=io_options,
                        )
                    return (problem_filename,), None
            else:
                raise NotImplementedError(
                    "There is currently no "
                    "script conversion available from "
                    "Pyomo to OSiL format."
                )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr import identify_variables, value, differentiate
import logging
import math

logger = logging.getLogger(__name__)


def _loop(derivs, e_vars, diff_mode, ndx_list):
    for ndx, item in enumerate(derivs):
        ndx_list.append(ndx)
        if item.__class__ == list:
            for a, b in _loop(item, e_vars, diff_mode, ndx_list):
                yield a, b
        else:
            _derivs = differentiate(item, wrt_list=e_vars, mode=diff_mode)
            derivs[ndx] = _derivs
            yield ndx_list, _derivs
        ndx_list.pop()


def taylor_series_expansion(
    expr, diff_mode=differentiate.Modes.reverse_numeric, order=1
):
    """
    Generate a taylor series approximation for expr.

    Parameters
    ----------
    expr: pyomo.core.expr.numeric_expr.NumericExpression
    diff_mode: pyomo.core.expr.calculus.derivatives.Modes
        The method for differentiation.
    order: The order of the taylor series expansion
        If order is not 1, then symbolic differentiation must
        be used (differentiation.Modes.reverse_symbolic or
        differentiation.Modes.sympy).

    Returns
    -------
    res: pyomo.core.expr.numeric_expr.NumericExpression
    """
    if order < 0:
        raise ValueError(
            'Cannot compute taylor series expansion of order {0}'.format(str(order))
        )
    if order != 1 and diff_mode is differentiate.Modes.reverse_numeric:
        logger.warning(
            'taylor_series_expansion can only use symbolic differentiation for orders larger than 1'
        )
        diff_mode = differentiate.Modes.reverse_symbolic
    e_vars = list(identify_variables(expr=expr, include_fixed=False))

    res = value(expr)
    if order >= 1:
        derivs = differentiate(expr=expr, wrt_list=e_vars, mode=diff_mode)
        res += sum(
            (e_vars[i] - e_vars[i].value) * value(derivs[i]) for i in range(len(e_vars))
        )

    """
    This last bit of code is just for higher order taylor series expansions.
    The recursive function _loop modifies derivs in place so that derivs becomes a 
    list of lists of lists... However, _loop is also a generator so that 
    we don't have to loop through it twice. _loop yields two lists. The 
    first is a list of indices corresponding to the first k-1 variables that
    differentiation is being done with respect to. The second is a list of 
    derivatives. Each entry in this list is the derivative with respect to 
    the first k-1 variables and the kth variable, whose index matches the 
    index in _derivs.
    """
    if order >= 2:
        for n in range(2, order + 1):
            coef = 1.0 / math.factorial(n)
            for ndx_list, _derivs in _loop(derivs, e_vars, diff_mode, list()):
                tmp = coef
                for ndx in ndx_list:
                    tmp *= e_vars[ndx] - e_vars[ndx].value
                res += tmp * sum(
                    (e_vars[i] - e_vars[i].value) * value(_derivs[i])
                    for i in range(len(e_vars))
                )

    return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

import pyomo.environ as pyo
from pyomo.gdp import Disjunct, Disjunction
from pyomo.gdp.util import check_model_algebraic
from pyomo.common.log import LoggingIntercept
import logging
from io import StringIO


class TestGDPReclassificationError(unittest.TestCase):
    def test_disjunct_not_in_disjunction(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.d1 = Disjunct()
        m.d1.c = pyo.Constraint(expr=m.x == 1)
        m.d2 = Disjunct()
        m.d2.c = pyo.Constraint(expr=m.x == 0)

        pyo.TransformationFactory('gdp.bigm').apply_to(m)
        log = StringIO()
        with LoggingIntercept(log, 'pyomo.gdp', logging.WARNING):
            check_model_algebraic(m)
        self.assertRegex(log.getvalue(), '.*not found in any Disjunctions.*')

    def test_disjunct_not_in_active_disjunction(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.d1 = Disjunct()
        m.d1.c = pyo.Constraint(expr=m.x == 1)
        m.d2 = Disjunct()
        m.d2.c = pyo.Constraint(expr=m.x == 0)
        m.disjunction = Disjunction(expr=[m.d1, m.d2])
        m.disjunction.deactivate()
        pyo.TransformationFactory('gdp.bigm').apply_to(m)
        log = StringIO()
        with LoggingIntercept(log, 'pyomo.gdp', logging.WARNING):
            check_model_algebraic(m)
        self.assertRegex(
            log.getvalue(),
            '.*While it participates in a Disjunction, '
            'that Disjunction is currently deactivated.*',
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# declare deprecation paths for removed modules
#
from pyomo.common.deprecation import relocated_module_attribute, moved_module

moved_module(
    'pyomo.contrib.parmest.ipopt_solver_wrapper',
    'pyomo.contrib.parmest.utils.ipopt_solver_wrapper',
    version='6.4.2',
)
relocated_module_attribute(
    'create_ef', 'pyomo.contrib.parmest.utils.create_ef', version='6.4.2'
)
relocated_module_attribute(
    'ipopt_solver_wrapper',
    'pyomo.contrib.parmest.utils.ipopt_solver_wrapper',
    version='6.4.2',
)
relocated_module_attribute(
    'mpi_utils', 'pyomo.contrib.parmest.utils.mpi_utils', version='6.4.2'
)
relocated_module_attribute(
    'scenario_tree', 'pyomo.contrib.parmest.utils.scenario_tree', version='6.4.2'
)
del relocated_module_attribute, moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

import logging

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.trustregion.util import minIgnoreNone, maxIgnoreNone
from pyomo.core import (
    Block,
    Param,
    VarList,
    Constraint,
    Objective,
    value,
    Set,
    ExternalFunction,
    maximize,
    minimize,
)
from pyomo.core.expr.calculus.derivatives import differentiate
from pyomo.core.expr.visitor import identify_variables, ExpressionReplacementVisitor
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression
from pyomo.core.expr.numvalue import native_types
from pyomo.opt import SolverFactory, check_optimal_termination

logger = logging.getLogger('pyomo.contrib.trustregion')


class EFReplacement(ExpressionReplacementVisitor):
    """
    This class is a subclass of ExpressionReplacementVisitor.
    It replaces an external function expression in an expression tree with a
    "holder" variable (recorded in a ComponentMap) and sets the initial value
    of the new node on the tree to that of the original node, if it can.

    NOTE: We use an empty substitution map. The EFs to be substituted are
          identified as part of exitNode.
    """

    def __init__(self, trfData, efSet):
        super().__init__(
            descend_into_named_expressions=True, remove_named_expressions=False
        )
        self.trfData = trfData
        self.efSet = efSet

    def beforeChild(self, node, child, child_idx):
        # We want to capture all of the variables on the model.
        # If we reject a step, we need to know all the vars to reset.
        descend, result = super().beforeChild(node, child, child_idx)
        if (
            not descend
            and result.__class__ not in native_types
            and result.is_variable_type()
        ):
            self.trfData.all_variables.add(result)
        return descend, result

    def exitNode(self, node, data):
        # This is where the replacement happens.
        new_node = super().exitNode(node, data)
        if new_node.__class__ is not ExternalFunctionExpression:
            return new_node
        if self.efSet is not None and new_node._fcn not in self.efSet:
            # SKIP: efSet is provided and this node is not in it.
            return new_node

        _output = self.trfData.ef_outputs.add()
        # Set the value for the new node to the evaluated value of the
        # original node, if possible.
        try:
            _output.set_value(value(node))
        except:
            _output.set_value(0)
        # Preserve the new node as a truth model.
        # self.TRF.truth_models is a ComponentMap.
        self.trfData.truth_models[_output] = new_node
        return _output


class TRFInterface:
    """
    Pyomo interface for Trust Region algorithm.
    """

    def __init__(self, model, decision_variables, ext_fcn_surrogate_map_rule, config):
        self.original_model = model
        tmp_name = unique_component_name(self.original_model, 'tmp')
        setattr(self.original_model, tmp_name, decision_variables)
        self.config = config
        self.model = self.original_model.clone()
        self.decision_variables = getattr(self.model, tmp_name)
        delattr(self.original_model, tmp_name)
        self.data = Block()
        self.model.add_component(
            unique_component_name(self.model, 'trf_data'), self.data
        )
        self.basis_expression_rule = ext_fcn_surrogate_map_rule
        self.efSet = None
        self.solver = SolverFactory(self.config.solver)
        # TODO: Provide an API for users to set this only to substitute
        # a subset of identified external functions.
        # Also rename to "efFilterSet" or something similar.

    def replaceEF(self, expr):
        """
        Replace an External Function.

        Arguments:
            expr  : a Pyomo expression. We will search this expression tree

        This function returns an expression after removing any
        ExternalFunction in the set efSet from the expression tree
        `expr` and replacing them with variables.
        New variables are declared on the `TRF` block.

        TODO: Future work - investigate direct substitution of basis or
        surrogate models using Expression objects instead of new variables.
        """
        return EFReplacement(self.data, self.efSet).walk_expression(expr)

    def _remove_ef_from_expr(self, component):
        """
        This method takes a component and looks at its expression.
        If the expression contains an external function (EF), a new expression
        with the EF replaced with a "holder" variable is added to the component
        and the basis expression for the new "holder" variable is updated.
        """
        expr = component.expr
        next_ef_id = len(self.data.ef_outputs)
        new_expr = self.replaceEF(expr)
        if new_expr is not expr:
            component.set_value(new_expr)
            new_output_vars = list(
                self.data.ef_outputs[i + 1]
                for i in range(next_ef_id, len(self.data.ef_outputs))
            )
            for v in new_output_vars:
                self.data.basis_expressions[v] = self.basis_expression_rule(
                    component, self.data.truth_models[v]
                )

    def replaceExternalFunctionsWithVariables(self):
        """
        This method sets up essential data objects on the new trf_data block
        on the model as well as triggers the replacement of external functions
        in expressions trees.

        Data objects created:
            self.data.all_variables : ComponentSet
                A set of all variables on the model, including "holder"
                variables from the EF replacement
            self.data.truth_models : ComponentMap
                A component map for replaced nodes that keeps track of
                the truth model for that replacement.
            self.data.basis_expressions : ComponentMap
                A component map for the Pyomo expressions for basis functions
                as they apply to each variable
            self.data.ef_inputs : Dict
                A dictionary that tracks the input variables for each EF
            self.data.ef_outputs : VarList
                A list of the "holder" variables which replaced the original
                External Function expressions
        """
        self.data.all_variables = ComponentSet()
        self.data.truth_models = ComponentMap()
        self.data.basis_expressions = ComponentMap()
        self.data.ef_inputs = {}
        self.data.ef_outputs = VarList()

        number_of_equality_constraints = 0
        for con in self.model.component_data_objects(Constraint, active=True):
            if con.lb == con.ub and con.lb is not None:
                number_of_equality_constraints += 1
            self._remove_ef_from_expr(con)

        self.degrees_of_freedom = (
            len(list(self.data.all_variables)) - number_of_equality_constraints
        )
        if self.degrees_of_freedom != len(self.decision_variables):
            raise ValueError(
                "replaceExternalFunctionsWithVariables: "
                "The degrees of freedom %d do not match the number of decision "
                "variables supplied %d."
                % (self.degrees_of_freedom, len(self.decision_variables))
            )

        for var in self.decision_variables:
            if var not in self.data.all_variables:
                raise ValueError(
                    "replaceExternalFunctionsWithVariables: "
                    f"The supplied decision variable {var.name} cannot "
                    "be found in the model variables."
                )

        self.data.objs = list(self.model.component_data_objects(Objective, active=True))
        # HACK: This is a hack that we will want to remove once the NL writer
        # has been corrected to not send unused EFs to the solver
        for ef in self.model.component_objects(ExternalFunction):
            ef.parent_block().del_component(ef)

        if len(self.data.objs) != 1:
            raise ValueError(
                "replaceExternalFunctionsWithVariables: "
                "TrustRegion only supports models with a single active Objective."
            )
        if self.data.objs[0].sense == maximize:
            self.data.objs[0].expr = -1 * self.data.objs[0].expr
            self.data.objs[0].sense = minimize
        self._remove_ef_from_expr(self.data.objs[0])

        for i in self.data.ef_outputs:
            self.data.ef_inputs[i] = list(
                identify_variables(
                    self.data.truth_models[self.data.ef_outputs[i]], include_fixed=False
                )
            )
        self.data.all_variables.update(self.data.ef_outputs.values())
        self.data.all_variables = list(self.data.all_variables)

    def createConstraints(self):
        """
        Create the basis constraint y = b(w) (equation 3) and the
        surrogate model constraint y = r_k(w) (equation 5)

        Both constraints are immediately deactivated after creation and
        are activated later as necessary.
        """
        b = self.data

        # This implements: y = b(w) from Yoshio/Biegler (2020)
        @b.Constraint(b.ef_outputs.index_set())
        def basis_constraint(b, i):
            ef_output_var = b.ef_outputs[i]
            return ef_output_var == b.basis_expressions[ef_output_var]

        b.basis_constraint.deactivate()

        b.INPUT_OUTPUT = Set(
            initialize=(
                (i, j)
                for i in b.ef_outputs.index_set()
                for j in range(len(b.ef_inputs[i]))
            )
        )
        b.basis_model_output = Param(b.ef_outputs.index_set(), mutable=True)
        b.grad_basis_model_output = Param(b.INPUT_OUTPUT, mutable=True)
        b.truth_model_output = Param(b.ef_outputs.index_set(), mutable=True)
        b.grad_truth_model_output = Param(b.INPUT_OUTPUT, mutable=True)
        b.value_of_ef_inputs = Param(b.INPUT_OUTPUT, mutable=True)

        # This implements: y = r_k(w)
        @b.Constraint(b.ef_outputs.index_set())
        def sm_constraint_basis(b, i):
            ef_output_var = b.ef_outputs[i]
            return ef_output_var == b.basis_expressions[
                ef_output_var
            ] + b.truth_model_output[i] - b.basis_model_output[i] + sum(
                (b.grad_truth_model_output[i, j] - b.grad_basis_model_output[i, j])
                * (w - b.value_of_ef_inputs[i, j])
                for j, w in enumerate(b.ef_inputs[i])
            )

        b.sm_constraint_basis.deactivate()

    def getCurrentDecisionVariableValues(self):
        """
        Return current decision variable values
        """
        decision_values = {}
        for var in self.decision_variables:
            decision_values[var.name] = value(var)
        return decision_values

    def updateDecisionVariableBounds(self, radius):
        """Update the TRSP_k decision variable bounds

        This corresponds to:

        .. math::
            || E^{-1} (u - u_k) || <= trust_radius

        We omit :math:`E^{-1}` because we assume that the users have
        correctly scaled their variables.

        """
        for var in self.decision_variables:
            var.setlb(
                maxIgnoreNone(
                    value(var) - radius, self.initial_decision_bounds[var.name][0]
                )
            )
            var.setub(
                minIgnoreNone(
                    value(var) + radius, self.initial_decision_bounds[var.name][1]
                )
            )

    def updateSurrogateModel(self):
        """
        The parameters needed for the surrogate model are the values of:
            b(w_k)      : basis_model_output
            d(w_k)      : truth_model_output
            grad b(w_k) : grad_basis_model_output
            grad d(w_k) : grad_truth_model_output
        """
        b = self.data
        for i, y in b.ef_outputs.items():
            b.basis_model_output[i] = value(b.basis_expressions[y])
            b.truth_model_output[i] = value(b.truth_models[y])
            # Basis functions are Pyomo expressions (in theory)
            gradBasis = differentiate(b.basis_expressions[y], wrt_list=b.ef_inputs[i])
            # These, however, are external functions
            gradTruth = differentiate(b.truth_models[y], wrt_list=b.ef_inputs[i])
            for j, w in enumerate(b.ef_inputs[i]):
                b.grad_basis_model_output[i, j] = gradBasis[j]
                b.grad_truth_model_output[i, j] = gradTruth[j]
                b.value_of_ef_inputs[i, j] = value(w)

    def getCurrentModelState(self):
        """
        Return current state of all model variables.
        This is necessary if we need to reject a step and move backwards.
        """
        return list(value(v, exception=False) for v in self.data.all_variables)

    def calculateFeasibility(self):
        """
        Feasibility measure (theta(x)) is:
            || y - d(w) ||_1
        """
        b = self.data
        return sum(
            abs(value(y) - value(b.truth_models[y])) for i, y in b.ef_outputs.items()
        )

    def calculateStepSizeInfNorm(self, original_values, new_values):
        """
        Taking original and new values, calculate the step-size norm ||s_k||:
            || u - u_k ||_inf

        We assume that the user has correctly scaled their variables.
        """
        original_vals = []
        new_vals = []
        for var, val in original_values.items():
            original_vals.append(val)
            new_vals.append(new_values[var])
        return max([abs(new - old) for new, old in zip(new_vals, original_vals)])

    def initializeProblem(self):
        """
        Initializes appropriate constraints, values, etc. for TRF problem

        Returns
        -------
            objective_value : Initial objective
            feasibility     : Initial feasibility measure

        STEPS:
            1. Create and solve PMP (eq. 3) and set equal to "x_0"
            2. Evaluate d(w_0)
            3. Evaluate initial feasibility measure (theta(x_0))
            4. Create initial SM (difference btw. low + high fidelity models)

        """
        self.replaceExternalFunctionsWithVariables()
        self.initial_decision_bounds = {}
        for var in self.decision_variables:
            self.initial_decision_bounds[var.name] = [var.lb, var.ub]
        self.createConstraints()
        self.data.basis_constraint.activate()
        objective_value, _, _ = self.solveModel()
        self.data.basis_constraint.deactivate()
        self.updateSurrogateModel()
        feasibility = self.calculateFeasibility()
        self.data.sm_constraint_basis.activate()
        return objective_value, feasibility

    def solveModel(self):
        """
        Call the specified solver to solve the problem.

        Returns
        -------
            self.data.objs[0] : Current objective value
            step_norm         : Current step size inf norm
            feasibility       : Current feasibility measure

        This also caches the previous values of the vars, just in case
        we need to access them later if a step is rejected
        """
        current_decision_values = self.getCurrentDecisionVariableValues()
        self.data.previous_model_state = self.getCurrentModelState()
        results = self.solver.solve(
            self.model, keepfiles=self.config.keepfiles, tee=self.config.tee
        )

        if not check_optimal_termination(results):
            raise ArithmeticError(
                'EXIT: Model solve failed with status {} and termination'
                ' condition(s) {}.'.format(
                    str(results.solver.status),
                    str(results.solver.termination_condition),
                )
            )

        self.model.solutions.load_from(results)
        new_decision_values = self.getCurrentDecisionVariableValues()
        step_norm = self.calculateStepSizeInfNorm(
            current_decision_values, new_decision_values
        )
        feasibility = self.calculateFeasibility()
        return self.data.objs[0](), step_norm, feasibility

    def rejectStep(self):
        """
        If a step is rejected, we reset the model variables values back
        to their cached state - which we set in solveModel
        """
        for var, val in zip(self.data.all_variables, self.data.previous_model_state):
            var.set_value(val, skip_validation=True)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from io import StringIO

from pyomo.common.log import LoggingIntercept


class SuppressConstantObjectiveWarning(LoggingIntercept):
    """Suppress the infeasible model warning message from solve().

    The "WARNING: Constant objective detected, replacing with a placeholder"
    warning message from calling solve() is often unwanted, but there is no
    clear way to suppress it.

    TODO need to fix this so that it only suppresses the desired message.

    """

    def __init__(self):
        super(SuppressConstantObjectiveWarning, self).__init__(
            StringIO(), 'pyomo.core', logging.WARNING
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

"""
This model is adapted from Noriyuki Yoshio's model for his and Biegler's
2021 publication in AIChE.


Yoshio, N, Biegler, L.T. Demand-based optimization of a chlorobenzene process
with high-fidelity and surrogate reactor models under trust region strategies.
AIChE J. 2021; 67:e17054. https://doi.org/10.1002/aic.17054
"""

from pyomo.environ import ConcreteModel, Var, ExternalFunction, Objective
from pyomo.opt import SolverFactory


def ext_fcn(a, b):
    return a**2 + b**2


def grad_ext_fcn(args, fixed):
    a, b = args[:2]
    return [2 * a, 2 * b]


def create_model():
    m = ConcreteModel()
    m.name = 'Example 2: Yoshio'

    m.x1 = Var(initialize=0)
    m.x2 = Var(bounds=(-2.0, None), initialize=0)

    m.EF = ExternalFunction(ext_fcn, grad_ext_fcn)

    @m.Constraint()
    def con(m):
        return 2 * m.x1 + m.x2 + 10.0 == m.EF(m.x1, m.x2)

    m.obj = Objective(expr=(m.x1 - 1) ** 2 + (m.x2 - 3) ** 2 + m.EF(m.x1, m.x2) ** 2)
    return m


def basis_rule(component, ef_expr):
    x = ef_expr.arg(0)
    y = ef_expr.arg(1)
    return x**2 - y  # This is the low fidelity model


# This problem takes more than the default maximum iterations (50) to solve.
# In testing (Mac 10.15/ipopt version 3.12.12 from conda),
# it took 70 iterations.
def main():
    m = create_model()
    optTRF = SolverFactory('trustregion', maximum_iterations=100, verbose=True)
    optTRF.solve(m, [m.x1], ext_fcn_surrogate_map_rule=basis_rule)


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from pyomo.common.log import is_debug_set
from pyomo.dataportal.factory import DataManagerFactory, UnknownDataManager

logger = logging.getLogger('pyomo.core')


class DataPortal:
    """
    An object that manages loading and storing data from external
    data sources.  This object interfaces to plugins that manipulate
    the data in a manner that is dependent on the data format.

    Internally, the data in a DataPortal object is organized as follows::

        data[namespace][symbol][index] -> value

    All data is associated with a symbol name, which may be indexed,
    and which may belong to a namespace.  The default namespace is
    :const:`None`.

    Args:
        model: The model for which this data is associated.  This is
            used for error checking (e.g. object names must
            exist in the model, set dimensions must match, etc.).
            Default is :const:`None`.
        filename (str): A file from which data is loaded.  Default
            is :const:`None`.
        data_dict (dict): A dictionary used to initialize the data
            in this object.  Default is :const:`None`.
    """

    def __init__(self, *args, **kwds):
        """
        Constructor
        """
        if len(args) > 0:
            raise RuntimeError(
                "Unexpected constructor argument for a DataPortal object"
            )

        # Initialize this object with no data manager
        self._data_manager = None

        # Map initialization data as follows: _data[namespace][symbol] -> data
        self._data = {}

        # This is the data that is imported from various sources
        self._default = {}

        # Get the model for which this data is associated.
        self._model = kwds.pop('model', None)

        # Load data from a file ...
        if 'filename' in kwds:
            filename = kwds.pop('filename')
            self.connect(filename=filename, **kwds)
            self.load()
            self.disconnect()
        # Or load data from a dictionary
        elif 'data_dict' in kwds:
            data = kwds.pop('data_dict')
            if not data is None:
                self._data = data
        elif len(kwds) > 0:
            raise ValueError("Unknown options: %s" % str(kwds.keys()))

    def connect(self, **kwds):
        """
        Construct a data manager object that is associated with the input source.
        This data manager is used to process future data imports and exports.

        Args:
            filename (str): A filename that specifies the data source.
                Default is :const:`None`.
            server (str): The name of the remote server that hosts the data.
                Default is :const:`None`.
            using (str): The name of the resource used to load the data.
                Default is :const:`None`.

        Other keyword arguments are passed to the data manager object.
        """
        if not self._data_manager is None:
            self._data_manager.close()
        data = kwds.get('using', None)
        if data is None:
            data = kwds.get('filename', None)
        if data is None:
            data = kwds.get('server', None)
        if '.' in data:
            tmp = data.split(".")[-1]
        else:
            tmp = data
        self._data_manager = DataManagerFactory(tmp)
        if type(self._data_manager) is UnknownDataManager:
            raise IOError("Unknown file format '%s'" % tmp)
        self._data_manager.initialize(**kwds)
        self._data_manager.open()

    def disconnect(self):
        """
        Close the data manager object that is associated with the
        input source.
        """
        self._data_manager.close()
        self._data_manager = None

    def load(self, **kwds):
        """
        Import data from an external data source.

        Args:
            model: The model object for which this data is associated.
                Default is :const:`None`.

        Other keyword arguments are passed to the :func:`connect()` method.
        """
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Loading data...")
        #
        # Process arguments
        #
        _model = kwds.pop('model', None)
        if not _model is None:
            self._model = _model
        #
        # If _disconnect is True, then disconnect the data
        # manager after we load data
        #
        _disconnect = False
        if self._data_manager is None:
            #
            # Start a new connection
            #
            self.connect(**kwds)
            _disconnect = True
        elif len(kwds) > 0:
            #
            # We are continuing to store using an existing connection.
            #
            # Q: Should we reinitialize?  The semantic difference between
            # initialize() and add_options() aren't clear.
            #
            self._data_manager.add_options(**kwds)
        #
        # Preprocess the command-line options
        #
        self._preprocess_options()
        #
        # Read from data manager into self._data and self._default
        #
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Processing data ...")
        self._data_manager.read()
        status = self._data_manager.process(self._model, self._data, self._default)
        self._data_manager.clear()
        #
        # Disconnect
        #
        if _disconnect:
            self.disconnect()
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Done.")

    def store(self, **kwds):
        """
        Export data to an external data source.

        Args:
            model: The model object for which this data is associated.
                Default is :const:`None`.

        Other keyword arguments are passed to the :func:`connect()` method.
        """
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Storing data...")
        #
        # Process arguments
        #
        _model = kwds.pop('model', None)
        if not _model is None:
            self._model = _model
        #
        # If _disconnect is True, then disconnect the data manager
        # after we load data
        #
        _disconnect = False
        if self._data_manager is None:
            self.connect(**kwds)
            _disconnect = True
        elif len(kwds) > 0:
            #
            # Q: Should we reinitialize?  The semantic difference between
            # initialize() and add_options() aren't clear.
            #
            self._data_manager.add_options(**kwds)
        #
        # Preprocess the command-line options
        #
        self._preprocess_options()
        self._load_data_from_model()
        #
        # Write from self._data
        #
        self._data_manager.write(self._data)
        #
        # Disconnect
        #
        if _disconnect:
            self.disconnect()
        if is_debug_set(logger):  # pragma:nocover
            logger.debug("Done.")

    def data(self, name=None, namespace=None):
        """
            Return the data associated with a symbol and namespace

        Args:
            name (str): The name of the symbol that is returned.
                Default is :const:`None`, which indicates that the
                entire data in the namespace is returned.
            namespace (str): The name of the namespace that is accessed.
                Default is :const:`None`.

        Returns:
            If ``name`` is :const:`None`, then the dictionary for
            the namespace is returned.  Otherwise, the data
            associated with ``name`` in given namespace is returned.
            The return value is a constant if :const:`None` if
            there is a single value in the symbol dictionary, and otherwise
            the symbol dictionary is returned.
        """
        if not namespace in self._data:
            raise IOError("Unknown namespace '%s'" % str(namespace))
        if name is None:
            return self._data[namespace]
        ans = self._data[namespace][name]
        if None in ans:
            # The data is a simple value
            return ans[None]
        return ans

    def __getitem__(self, *args):
        """
        Return the specified data value.

        If a single argument is given, then this is the symbol name::

            dp = DataPortal()
            dp[name]

        If a two arguments are given, then the first is the namespace and
        the second is the symbol name::

            dp = DataPortal()
            dp[namespace, name]

        Args:
            *args (str): A tuple of arguments.

        Returns:
            If a single argument is given, then the data associated
            with that symbol in the namespace :const:`None` is returned.
            If two arguments are given, then the data associated with
            symbol in the given namespace is returned.
        """
        if type(args[0]) is tuple or type(args[0]) is list:
            assert len(args) == 1
            args = args[0]
        if len(args) > 2:
            raise IOError(
                "Must specify data name:  DataPortal[name] or Data[namespace, name]"
            )
        elif len(args) == 2:
            namespace = args[0]
            name = args[1]
        else:
            namespace = None
            name = args[0]

        ans = self._data[namespace][name]
        if None in ans:
            # The data is a simple value
            return ans[None]
        return ans

    def __setitem__(self, name, value):
        """
        Set the value of ``name`` with the given value.

        Args:
            name (str): The name of the symbol that is set.
            value: The value of the symbol.
        """
        if not None in self._data:
            self._data[None] = {}
        self._data[None][name] = value

    def namespaces(self):
        """
        Return an iterator for the namespaces in the data portal.

        Yields:
            A string name for the next namespace.
        """
        for key in self._data:
            yield key

    def keys(self, namespace=None):
        """
        Return an iterator of the data keys in
        the specified namespace.

        Yields:
            A string name for the next symbol in the specified namespace.
        """
        for key in self._data[namespace]:
            yield key

    def values(self, namespace=None):
        """
        Return an iterator of the data values in
        the specified namespace.

        Yields:
            The data value for the next symbol in the specified namespace.
            This may be a simple value, or a dictionary of values.
        """
        for key in self._data[namespace]:
            ans = self._data[namespace][key]
            if None in ans:
                yield ans[None]
            else:
                yield ans

    def items(self, namespace=None):
        """
        Return an iterator of (name, value) tuples from the data in
        the specified namespace.

        Yields:
            The next (name, value) tuple in the namespace.  If the symbol
            has a simple data value, then that is included in the tuple.
            Otherwise, the tuple includes a dictionary mapping
            symbol indices to values.
        """
        for key in self._data[namespace]:
            ans = self._data[namespace][key]
            if None in ans:
                yield key, ans[None]
            else:
                yield key, ans

    def _preprocess_options(self):
        """
        Preprocess the options for a data manager.
        """
        options = self._data_manager.options
        #
        if options.data is None and (
            not options.set is None
            or not options.param is None
            or not options.index is None
        ):
            #
            # Set options.data to a list of elements of the options.set,
            # options.param and options.index values.
            #
            options.data = []
            if not options.set is None:
                assert type(options.set) not in (list, tuple)
                options.data.append(options.set)
                #
                # The set option should not be a list or tuple.
                #
                # if type(options.set) in (list,tuple):
                #    for item in options.set:
                #        options.data.append(item)
                # else:
                #    options.data.append(options.set)
            if not options.index is None:
                options.data.append(options.index)
            if not options.param is None:
                if type(options.param) in (list, tuple):
                    for item in options.param:
                        options.data.append(item)
                else:
                    options.data.append(options.param)
        #
        if options.data is None:
            return
        #
        if type(options.data) in (list, tuple):
            #
            # If options.data is a list/tuple, then
            # process it to get the names of the
            # elements.  Thus, if a component is included
            # in options.data, then it is replaced by its name.
            #
            ans = []
            for item in options.data:
                try:
                    ans.append(item.local_name)
                    self._model = item.model()
                except:
                    ans.append(item)
            options.data = ans
        else:
            #
            # If options.data is a single value, then we assume that
            # it is a component.  Reset its value to the value of
            # the component name.
            #
            try:
                self._model = options.data.model()
                options.data = [self._data_manager.options.data.local_name]
            except:
                pass

    def _load_data_from_model(self):
        """
        Load model data into self._data
        """
        if self._data_manager.options.data is None or self._model is None:
            return
        for name in self._data_manager.options.data:
            c = getattr(self._model, name)
            try:
                self._data[name] = c.data()
            except:
                self._data[name] = c.extract_values()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import itertools
from enum import Enum
from pyomo.common.errors import DeveloperError
from pyomo.common.dependencies import numpy as np
from pyomo.contrib.piecewise.ordered_3d_j1_triangulation_data import (
    get_hamiltonian_paths,
)


class Triangulation(Enum):
    Unknown = 0
    AssumeValid = 1
    Delaunay = 2
    J1 = 3
    OrderedJ1 = 4


# Duck-typed thing that looks reasonably similar to an instance of
# scipy.spatial.Delaunay
# Fields:
#   - points: list of P points as P x n array
#   - simplices: list of M simplices as P x (n + 1) array of point _indices_
#   - coplanar: list of N points omitted from triangulation as tuples of (point index,
#     nearest simplex index, nearest vertex index), stacked into an N x 3 array
class _Triangulation:
    def __init__(self, points, simplices, coplanar):
        self.points = points
        self.simplices = simplices
        self.coplanar = coplanar


# Get an unordered J1 triangulation, as described by [1], of a finite grid of
# points in R^n having the same odd number of points along each axis.
# References
# ----------
# [1] J.P. Vielma, S. Ahmed, and G. Nemhauser, "Mixed-integer models
#     for nonseparable piecewise-linear optimization: unifying framework
#     and extensions," Operations Research, vol. 58, no. 2, pp. 305-315,
#     2010.
def get_unordered_j1_triangulation(points, dimension):
    points_map, num_pts = _process_points_j1(points, dimension)
    simplices_list = _get_j1_triangulation(points_map, num_pts - 1, dimension)
    return _Triangulation(
        points=np.array(points),
        simplices=np.array(simplices_list),
        coplanar=np.array([]),
    )


# Get an ordered J1 triangulation, according to [1], with the additional condition
# added from [2] that simplex vertices are also ordered such that the final vertex
# of each simplex is the first vertex of the next simplex.
# References
# ----------
# [1] Michael J. Todd. "Hamiltonian triangulations of Rn". In: Functional
#     Differential Equations and Approximation of Fixed Points. Ed. by
#     Heinz-Otto Peitgen and Hans-Otto Walther. Berlin, Heidelberg: Springer
#     Berlin Heidelberg, 1979, pp. 470–483. ISBN: 978-3-540-35129-0.
# [2] J.P. Vielma, S. Ahmed, and G. Nemhauser, "Mixed-integer models
#     for nonseparable piecewise-linear optimization: unifying framework
#     and extensions," Operations Research, vol. 58, no. 2, pp. 305-315,
#     2010.
def get_ordered_j1_triangulation(points, dimension):
    points_map, num_pts = _process_points_j1(points, dimension)
    if dimension == 2:
        simplices_list = _get_ordered_j1_triangulation_2d(points_map, num_pts - 1)
    elif dimension == 3:
        simplices_list = _get_ordered_j1_triangulation_3d(points_map, num_pts - 1)
    else:
        simplices_list = _get_ordered_j1_triangulation_4d_and_above(
            points_map, num_pts - 1, dimension
        )
    return _Triangulation(
        points=np.array(points),
        simplices=np.array(simplices_list),
        coplanar=np.array([]),
    )


# Does some validation but mostly assumes the user did the right thing
def _process_points_j1(points, dimension):
    if not len(points[0]) == dimension:
        raise ValueError("Points not consistent with specified dimension")
    num_pts = round(len(points) ** (1 / dimension))
    if not len(points) == num_pts**dimension:
        raise ValueError(
            "'points' must have points forming an n-dimensional grid with straight grid"
            " lines and the same odd number of points in each axis."
        )
    if not num_pts % 2 == 1:
        raise ValueError(
            "'points' must have points forming an n-dimensional grid with straight grid"
            " lines and the same odd number of points in each axis."
        )

    # munge the points into an organized map from n-dimensional keys to original
    # indices
    points.sort()
    points_map = {}
    for point_index in itertools.product(range(num_pts), repeat=dimension):
        point_flat_index = 0
        for n in range(dimension):
            point_flat_index += point_index[dimension - 1 - n] * num_pts**n
        points_map[point_index] = point_flat_index
    return points_map, num_pts


# Implement the J1 "Union Jack" triangulation (Todd 79) as explained by
# Vielma 2010, with no ordering guarantees imposed. This function triangulates
# {0, ..., K}^n for even K using the J1 triangulation, mapping the
# obtained simplices through the points_map for a slight generalization.
def _get_j1_triangulation(points_map, K, n):
    if K % 2 != 0:
        raise ValueError("K must be even")
    # 1, 3, ..., K - 1
    axis_odds = range(1, K, 2)
    V_0 = itertools.product(axis_odds, repeat=n)
    big_iterator = itertools.product(
        V_0,
        itertools.permutations(range(0, n), n),
        itertools.product((-1, 1), repeat=n),
    )
    ret = []
    for v_0, pi, s in big_iterator:
        simplex = []
        current = list(v_0)
        simplex.append(points_map[tuple(current)])
        for i in range(0, n):
            current[pi[i]] += s[pi[i]]
            simplex.append(points_map[tuple(current)])
        # sort this because it might happen again later and we'd like to stay
        # consistent. Undo this if it's slow.
        ret.append(sorted(simplex))
    return ret


class Direction(Enum):
    left = 0
    down = 1
    up = 2
    right = 3


# Implement something similar to proof-by-picture from Todd 79 (Figure 1).
# However, that drawing is misleading at best so I do it in a working way, and
# also slightly more regularly. I also go from the outside in instead of from
# the inside out, to make things easier to implement.
def _get_ordered_j1_triangulation_2d(points_map, num_pts):
    # check when square has simplices in top-left and bottom-right
    square_parity_tlbr = lambda x, y: x % 2 == y % 2
    # check when we are in a "turnaround square" as seen in the picture
    is_turnaround = lambda x, y: x >= num_pts / 2 and y == (num_pts / 2) - 1

    facing = None

    simplices = []
    start_square = (num_pts - 1, (num_pts / 2) - 1)

    # make it easier to read what I'm doing
    def add_bottom_right():
        simplices.append(
            (points_map[x, y], points_map[x + 1, y], points_map[x + 1, y + 1])
        )

    def add_top_right():
        simplices.append(
            (points_map[x, y + 1], points_map[x + 1, y], points_map[x + 1, y + 1])
        )

    def add_bottom_left():
        simplices.append((points_map[x, y], points_map[x, y + 1], points_map[x + 1, y]))

    def add_top_left():
        simplices.append(
            (points_map[x, y], points_map[x, y + 1], points_map[x + 1, y + 1])
        )

    # identify square by bottom-left corner
    x, y = start_square
    used_squares = set()  # not used for the turnaround squares

    # depending on parity we will need to go either up or down to start
    if square_parity_tlbr(x, y):
        add_bottom_right()
        facing = Direction.down
        y -= 1
    else:
        add_top_right()
        facing = Direction.up
        y += 1

    # state machine
    while True:
        if facing == Direction.left:
            if square_parity_tlbr(x, y):
                add_bottom_right()
                add_top_left()
            else:
                add_top_right()
                add_bottom_left()
            used_squares.add((x, y))
            if (x - 1, y) in used_squares or x == 0:
                # can't keep going left so we need to go up or down depending
                # on parity
                if square_parity_tlbr(x, y):
                    y += 1
                    facing = Direction.up
                    continue
                else:
                    y -= 1
                    facing = Direction.down
                    continue
            else:
                x -= 1
                continue
        elif facing == Direction.right:
            if is_turnaround(x, y):
                # finished; this case should always eventually be reached
                add_bottom_left()
                _fix_vertices_incremental_order(simplices)
                return simplices
            else:
                if square_parity_tlbr(x, y):
                    add_top_left()
                    add_bottom_right()
                else:
                    add_bottom_left()
                    add_top_right()
            used_squares.add((x, y))
            if (x + 1, y) in used_squares or x == num_pts - 1:
                # can't keep going right so we need to go up or down depending
                # on parity
                if square_parity_tlbr(x, y):
                    y -= 1
                    facing = Direction.down
                    continue
                else:
                    y += 1
                    facing = Direction.up
                    continue
            else:
                x += 1
                continue
        elif facing == Direction.down:
            if is_turnaround(x, y):
                # we are always in a TLBR square. Take the TL of this, the TR
                # of the one on the left, and continue upwards one to the left
                add_top_left()
                x -= 1
                add_top_right()
                y += 1
                facing = Direction.up
                continue
            else:
                if square_parity_tlbr(x, y):
                    add_top_left()
                    add_bottom_right()
                else:
                    add_top_right()
                    add_bottom_left()
                used_squares.add((x, y))
                if (x, y - 1) in used_squares or y == 0:
                    # can't keep going down so we need to turn depending
                    # on our parity
                    if square_parity_tlbr(x, y):
                        x += 1
                        facing = Direction.right
                        continue
                    else:
                        x -= 1
                        facing = Direction.left
                        continue
                else:
                    y -= 1
                    continue
        elif facing == Direction.up:
            if is_turnaround(x, y):
                # we are always in a non-TLBR square. Take the BL of this, the BR
                # of the one on the left, and continue downwards one to the left
                add_bottom_left()
                x -= 1
                add_bottom_right()
                y -= 1
                facing = Direction.down
                continue
            else:
                if square_parity_tlbr(x, y):
                    add_bottom_right()
                    add_top_left()
                else:
                    add_bottom_left()
                    add_top_right()
                used_squares.add((x, y))
                if (x, y + 1) in used_squares or y == num_pts - 1:
                    # can't keep going up so we need to turn depending
                    # on our parity
                    if square_parity_tlbr(x, y):
                        x -= 1
                        facing = Direction.left
                        continue
                    else:
                        x += 1
                        facing = Direction.right
                        continue
                else:
                    y += 1
                    continue


def _get_ordered_j1_triangulation_3d(points_map, num_pts):
    incremental_3d_simplex_pair_to_path = get_hamiltonian_paths()
    # To start, we need a hamiltonian path in the grid graph of *double* cubes
    # (2x2x2 cubes)
    grid_hamiltonian = _get_grid_hamiltonian(3, round(num_pts / 2))  # division is exact

    # We always start by going from [0, 0, 0] to [0, 0, 1], so we can safely
    # start from the -x side.
    # Data format: the first tuple is a basis vector or its negative, representing a
    # face. The number afterwards is a 1 or 2 disambiguating which, of the two simplices
    # on that face we consider, we are referring to.
    start_data = ((-1, 0, 0), 1)

    simplices = []
    for i in range(len(grid_hamiltonian) - 1):
        current_double_cube_idx = grid_hamiltonian[i]
        next_double_cube_idx = grid_hamiltonian[i + 1]
        direction_to_next = tuple(
            next_double_cube_idx[j] - current_double_cube_idx[j] for j in range(3)
        )

        current_v_0 = tuple(2 * current_double_cube_idx[j] + 1 for j in range(3))

        current_cube_path = None
        if (
            start_data,
            (direction_to_next, 1),
        ) in incremental_3d_simplex_pair_to_path.keys():
            current_cube_path = incremental_3d_simplex_pair_to_path[
                (start_data, (direction_to_next, 1))
            ]
            # set the start data for the next iteration now
            start_data = (tuple(-1 * i for i in direction_to_next), 1)
        else:
            current_cube_path = incremental_3d_simplex_pair_to_path[
                (start_data, (direction_to_next, 2))
            ]
            start_data = (tuple(-1 * i for i in direction_to_next), 2)

        for simplex_data in current_cube_path:
            simplices.append(
                _get_one_j1_simplex(
                    current_v_0, simplex_data[1], simplex_data[0], 3, points_map
                )
            )

    # fill in the last cube. We have a good start_data but we need to invent a
    # direction_to_next. Let's go straight in the direction we came from.
    direction_to_next = tuple(-1 * i for i in start_data[0])
    current_v_0 = tuple(2 * grid_hamiltonian[-1][j] + 1 for j in range(3))
    if (
        start_data,
        (direction_to_next, 1),
    ) in incremental_3d_simplex_pair_to_path.keys():
        current_cube_path = incremental_3d_simplex_pair_to_path[
            (start_data, (direction_to_next, 1))
        ]
    else:
        current_cube_path = incremental_3d_simplex_pair_to_path[
            (start_data, (direction_to_next, 2))
        ]

    for simplex_data in current_cube_path:
        simplices.append(
            _get_one_j1_simplex(
                current_v_0, simplex_data[1], simplex_data[0], 3, points_map
            )
        )

    _fix_vertices_incremental_order(simplices)
    return simplices


def _get_ordered_j1_triangulation_4d_and_above(points_map, num_pts, dim):
    # step one: get a hamiltonian path in the appropriate grid graph (low-coordinate
    # corners of the grid squares)
    grid_hamiltonian = _get_grid_hamiltonian(dim, num_pts)

    # step 1.5: get a starting simplex. Anything that is *not* adjacent to the
    # second square is fine. Since we always go from [0, ..., 0] to [0, ..., 1],
    # i.e., j=`dim`, anything where `dim` is not the first or last symbol should
    # always work. Let's stick it in the second place
    start_perm = tuple([1] + [dim] + list(range(2, dim)))

    # step two: for each square, get a sequence of simplices from a starting simplex,
    # through the square, and then ending with a simplex adjacent to the next square.
    # Then find the appropriate adjacent simplex to start on the next square
    simplices = []
    for i in range(len(grid_hamiltonian) - 1):
        current_corner = grid_hamiltonian[i]
        next_corner = grid_hamiltonian[i + 1]
        # differing index
        j = [k + 1 for k in range(dim) if current_corner[k] != next_corner[k]][0]
        # border x_j value between this square and next
        c = max(current_corner[j - 1], next_corner[j - 1])
        v_0, sign = _get_nearest_odd_and_sign_vec(current_corner)
        # According to Todd, what we need is to end with a permutation where rho(n) = j
        # if c is odd, and end with one where rho(1) = j if c is even. I think this
        # is right -- basically the sign from the sign vector sometimes cancels
        # out the sign from whether we are entering in the +c or -c direction.
        if c % 2 == 0:
            perm_sequence = _get_Gn_hamiltonian(dim, start_perm, j, False)
            for pi in perm_sequence:
                simplices.append(_get_one_j1_simplex(v_0, pi, sign, dim, points_map))
        else:
            perm_sequence = _get_Gn_hamiltonian(dim, start_perm, j, True)
            for pi in perm_sequence:
                simplices.append(_get_one_j1_simplex(v_0, pi, sign, dim, points_map))
        # should be true regardless of odd or even
        start_perm = perm_sequence[-1]

    # step three: finish out the last square
    # Any final permutation is fine; we are going nowhere after this
    v_0, sign = _get_nearest_odd_and_sign_vec(grid_hamiltonian[-1])
    for pi in _get_Gn_hamiltonian(dim, start_perm, 1, False):
        simplices.append(_get_one_j1_simplex(v_0, pi, sign, dim, points_map))

    # fix vertices and return
    _fix_vertices_incremental_order(simplices)
    return simplices


def _get_one_j1_simplex(v_0, pi, sign, dim, points_map):
    simplex = []
    current = list(v_0)
    simplex.append(points_map[tuple(current)])
    for i in range(0, dim):
        current[pi[i] - 1] += sign[pi[i] - 1]
        simplex.append(points_map[tuple(current)])
    return sorted(simplex)


# get the v_0 and sign vectors corresponding to a given square, identified by its
# low-coordinate corner
def _get_nearest_odd_and_sign_vec(corner):
    v_0 = []
    sign = []
    for x in corner:
        if x % 2 == 0:
            v_0.append(x + 1)
            sign.append(-1)
        else:
            v_0.append(x)
            sign.append(1)
    return v_0, sign


def _get_grid_hamiltonian(dim, length):
    if dim == 1:
        return [[n] for n in range(length)]
    else:
        ret = []
        prev = _get_grid_hamiltonian(dim - 1, length)
        for n in range(length):
            # if n is even, add the previous hamiltonian with n in its new first
            # coordinate. If odd, do the same with the previous hamiltonian in reverse.
            if n % 2 == 0:
                for x in prev:
                    ret.append([n] + x)
            else:
                for x in reversed(prev):
                    ret.append([n] + x)
        return ret


# Fix vertices (in place) when the simplices are right but vertices are not
def _fix_vertices_incremental_order(simplices):
    last_vertex_index = len(simplices[0]) - 1
    for i, simplex in enumerate(simplices):
        # Choose vertices like this: first is always the same as last
        # of the previous simplex. Last is arbitrarily chosen from the
        # intersection with the next simplex.
        first = None
        last = None
        if i == 0:
            first = 0
        else:
            first = simplex.index(simplices[i - 1][last_vertex_index])

        if i == len(simplices) - 1:
            last = last_vertex_index
        else:
            for n in range(last_vertex_index + 1):
                if simplex[n] in simplices[i + 1] and n != first:
                    last = n
                    break
            else:
                # For the Python neophytes in the audience (and other sane
                # people), the 'else' only runs if we do *not* break out of the
                # for loop.
                raise DeveloperError("Couldn't fix vertex ordering for incremental.")

        # reorder the simplex with the desired first and last
        new_simplex = list(simplex)
        temp = new_simplex[0]
        new_simplex[0] = new_simplex[first]
        new_simplex[first] = temp
        if last == 0:
            last = first
        temp = new_simplex[last_vertex_index]
        new_simplex[last_vertex_index] = new_simplex[last]
        new_simplex[last] = temp
        simplices[i] = tuple(new_simplex)


# Let G_n be the graph on n! vertices where the vertices are permutations in
# S_n and  two vertices are adjacent if they are related by swapping the values
# of pi(i - 1) and pi(i) for some i in {2, ..., n}.
#
# This function gets a Hamiltonian path through G_n, starting from a fixed
# starting permutation, such that a fixed target symbol is either the image
# rho(1), or it is rho(n), depending on whether first or last is requested,
# where rho is the final permutation.
def _get_Gn_hamiltonian(n, start_permutation, target_symbol, last, _cache={}):
    if n < 4:
        raise ValueError("n must be at least 4 for this operation to be possible")
    if (n, start_permutation, target_symbol, last) in _cache:
        return _cache[(n, start_permutation, target_symbol, last)]
    # first is enough because we can just reverse every permutation
    if last:
        ret = [
            tuple(reversed(pi))
            for pi in _get_Gn_hamiltonian(
                n, tuple(reversed(start_permutation)), target_symbol, False
            )
        ]
        _cache[(n, start_permutation, target_symbol, last)] = ret
        return ret
    # trivial start permutation is enough because we can map it through at the end
    if start_permutation != tuple(range(1, n + 1)):
        new_target_symbol = [
            x for x in range(1, n + 1) if start_permutation[x - 1] == target_symbol
        ][
            0
        ]  # pi^-1(j)
        ret = [
            tuple(start_permutation[pi[i] - 1] for i in range(n))
            for pi in _get_Gn_hamiltonian_impl(n, new_target_symbol)
        ]
        _cache[(n, start_permutation, target_symbol, last)] = ret
        return ret
    else:
        ret = _get_Gn_hamiltonian_impl(n, target_symbol)
        _cache[(n, start_permutation, target_symbol, last)] = ret
        return ret


# Assume the starting permutation is (1, ..., n) and the target symbol needs to
# be in the first position of the last permutation
def _get_Gn_hamiltonian_impl(n, target_symbol):
    # base case: proof by picture from Todd 79, Figure 2
    # note: Figure 2 contains an error, careful!
    if n == 4:
        if target_symbol == 1:
            return [
                (1, 2, 3, 4),
                (2, 1, 3, 4),
                (2, 1, 4, 3),
                (2, 4, 1, 3),
                (4, 2, 1, 3),
                (4, 2, 3, 1),
                (2, 4, 3, 1),
                (2, 3, 4, 1),
                (2, 3, 1, 4),
                (3, 2, 1, 4),
                (3, 2, 4, 1),
                (3, 4, 2, 1),
                (4, 3, 2, 1),
                (4, 3, 1, 2),
                (3, 4, 1, 2),
                (3, 1, 4, 2),
                (3, 1, 2, 4),
                (1, 3, 2, 4),
                (1, 3, 4, 2),
                (1, 4, 3, 2),
                (4, 1, 3, 2),
                (4, 1, 2, 3),
                (1, 4, 2, 3),
                (1, 2, 4, 3),
            ]
        elif target_symbol == 2:
            return [
                (1, 2, 3, 4),
                (1, 2, 4, 3),
                (1, 4, 2, 3),
                (4, 1, 2, 3),
                (4, 1, 3, 2),
                (1, 4, 3, 2),
                (1, 3, 4, 2),
                (1, 3, 2, 4),
                (3, 1, 2, 4),
                (3, 1, 4, 2),
                (3, 4, 1, 2),
                (4, 3, 1, 2),
                (4, 3, 2, 1),
                (3, 4, 2, 1),
                (3, 2, 4, 1),
                (3, 2, 1, 4),
                (2, 3, 1, 4),
                (2, 3, 4, 1),
                (2, 4, 3, 1),
                (4, 2, 3, 1),
                (4, 2, 1, 3),
                (2, 4, 1, 3),
                (2, 1, 4, 3),
                (2, 1, 3, 4),
            ]
        elif target_symbol == 3:
            return [
                (1, 2, 3, 4),
                (1, 2, 4, 3),
                (1, 4, 2, 3),
                (4, 1, 2, 3),
                (4, 1, 3, 2),
                (1, 4, 3, 2),
                (1, 3, 4, 2),
                (1, 3, 2, 4),
                (3, 1, 2, 4),
                (3, 1, 4, 2),
                (3, 4, 1, 2),
                (4, 3, 1, 2),
                (4, 3, 2, 1),
                (3, 4, 2, 1),
                (3, 2, 4, 1),
                (2, 3, 4, 1),
                (2, 4, 3, 1),
                (4, 2, 3, 1),
                (4, 2, 1, 3),
                (2, 4, 1, 3),
                (2, 1, 4, 3),
                (2, 1, 3, 4),
                (2, 3, 1, 4),
                (3, 2, 1, 4),
            ]
        elif target_symbol == 4:
            return [
                (1, 2, 3, 4),
                (2, 1, 3, 4),
                (2, 3, 1, 4),
                (3, 2, 1, 4),
                (3, 1, 2, 4),
                (1, 3, 2, 4),
                (1, 3, 4, 2),
                (3, 1, 4, 2),
                (3, 4, 1, 2),
                (3, 4, 2, 1),
                (3, 2, 4, 1),
                (2, 3, 4, 1),
                (2, 4, 3, 1),
                (2, 4, 1, 3),
                (2, 1, 4, 3),
                (1, 2, 4, 3),
                (1, 4, 2, 3),
                (1, 4, 3, 2),
                (4, 1, 3, 2),
                (4, 3, 1, 2),
                (4, 3, 2, 1),
                (4, 2, 3, 1),
                (4, 2, 1, 3),
                (4, 1, 2, 3),
            ]
        # unreachable
    else:
        # recursive case
        if target_symbol < n:  # Less awful case
            idx = n - 1
            facing = -1
            ret = []
            for pi in _get_Gn_hamiltonian_impl(n - 1, target_symbol):
                for _ in range(n):
                    l = list(pi)
                    l.insert(idx, n)
                    ret.append(tuple(l))
                    idx += facing
                    if idx == -1 or idx == n:  # went too far
                        facing *= -1
                        idx += facing  # stay once because we get a new pi
            return ret
        else:  # awful case, target_symbol = n
            idx = 0
            facing = 1
            ret = []
            for pi in _get_Gn_hamiltonian_impl(n - 1, n - 1):
                for _ in range(n):
                    l = [x + 1 for x in pi]
                    l.insert(idx, 1)
                    ret.append(tuple(l))
                    idx += facing
                    if idx == -1 or idx == n:  # went too far
                        facing *= -1
                        idx += facing  # stay once because we get a new pi
            # now we almost have a correct sequence, but it ends with (1, n, ...)
            # instead of (n, 1, ...) so we need to do some surgery
            last = ret.pop()  # of form (1, n, i, j, ...)
            second_last = ret.pop()  # of form (n, 1, i, j, ...)
            i = last[2]
            j = last[3]
            test = list(
                last
            )  # want permutation of form (n, 1, j, i, ...) with same tail
            test[0] = n
            test[1] = 1
            test[2] = j
            test[3] = i
            idx = ret.index(tuple(test))
            ret.insert(idx, second_last)
            ret.insert(idx, last)
            return ret



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    Set,
    ConstraintList,
    Block,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


def inactive_index_LP_obj_rule(model, i):
    if i == 1:
        return model.x - model.y
    else:
        return -model.x + model.y + model.z


def inactive_index_LP_c2_rule(model, i):
    if i == 1:
        return model.y >= -2
    else:
        return model.x <= 2


@register_model
class LP_inactive_index(_BaseTestModel):
    """
    A continuous linear model where component subindices have been deactivated
    """

    description = "LP_inactive_index"
    capabilities = set(['linear'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.s = Set(initialize=[1, 2])
        model.x = Var()
        model.y = Var()
        model.z = Var(bounds=(0, None))

        model.obj = Objective(model.s, rule=inactive_index_LP_obj_rule)
        model.OBJ = Objective(expr=model.x + model.y)
        model.obj[1].deactivate()
        model.OBJ.deactivate()
        model.c1 = ConstraintList()
        model.c1.add(model.x <= 1)  # index=1
        model.c1.add(model.x >= -1)  # index=2
        model.c1.add(model.y <= 1)  # index=3
        model.c1.add(model.y >= -1)  # index=4
        model.c1[1].deactivate()
        model.c1[4].deactivate()
        model.c2 = Constraint(model.s, rule=inactive_index_LP_c2_rule)

        model.b = Block()
        model.b.c = Constraint(expr=model.z >= 2)
        model.B = Block(model.s)
        model.B[1].c = Constraint(expr=model.z >= 3)
        model.B[2].c = Constraint(expr=model.z >= 1)

        model.b.deactivate()
        model.B.deactivate()
        model.B[2].activate()

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = None
        model.y.value = None
        model.z.value = 2.0


@register_model
class LP_inactive_index_kernel(LP_inactive_index):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.s = [1, 2]
        model.x = pmo.variable()
        model.y = pmo.variable()
        model.z = pmo.variable(lb=0)

        model.obj = pmo.objective_dict()
        for i in model.s:
            model.obj[i] = pmo.objective(inactive_index_LP_obj_rule(model, i))

        model.OBJ = pmo.objective(model.x + model.y)
        model.obj[1].deactivate()
        model.OBJ.deactivate()
        model.c1 = pmo.constraint_dict()
        model.c1[1] = pmo.constraint(model.x <= 1)
        model.c1[2] = pmo.constraint(model.x >= -1)
        model.c1[3] = pmo.constraint(model.y <= 1)
        model.c1[4] = pmo.constraint(model.y >= -1)
        model.c1[1].deactivate()
        model.c1[4].deactivate()
        model.c2 = pmo.constraint_dict()
        for i in model.s:
            model.c2[i] = pmo.constraint(inactive_index_LP_c2_rule(model, i))

        model.b = pmo.block()
        model.b.c = pmo.constraint(model.z >= 2)
        model.B = pmo.block_dict()
        model.B[1] = pmo.block()
        model.B[1].c = pmo.constraint(model.z >= 3)
        model.B[2] = pmo.block()
        model.B[2].c = pmo.constraint(model.z >= 1)

        model.b.deactivate()
        model.B[1].deactivate()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .scalar_data import ScalarData
from .series_data import TimeSeriesData
from .interval_data import IntervalData
from .convert import series_to_interval, interval_to_series

__doc__ = """A module containing data structures for storing values associated
    with time-indexed Pyomo variables.

    This is the core of the mpc package. Code in this module should not
    import from other parts of mpc.

    """
